#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field
from typing import Literal, Iterable, Mapping, ClassVar, Container

from frozendict import frozendict
from init_helpers import Jsonable

from .response import Response
from .security import SecurityRequirement
from .parameter import SpecParameter
from .request_body import RequestBody, SpecRef, SpecResource
from .tag import Tag
from .referencable_resource import ReferencableResource


@dataclass(frozen=True, kw_only=True, slots=True)
class Operation(ReferencableResource):
    """Operation https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#operation-object"""
    supported_methods: ClassVar[Container[str]] = ('get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace')

    path: str
    method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'HEAD', 'OPTIONS', 'TRACE',
                    'get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace']   # HTTP method
    operation_id: str  # MUST be unique among all operations in spec
    securities: Iterable[SecurityRequirement] = tuple()  # empty iterable means "public"(no limitations)
    parameters: Iterable[SpecParameter] = tuple()
    request_body: RequestBody | None = None
    # code_to_answer: Mapping[int | Literal['default'], type[Answer]] = field(default_factory=frozendict)
    code_to_response: Mapping[int | Literal['default'], Response] = field(default_factory=frozendict)
    summary: str = ''
    description: str = ''
    tags: Iterable[str | Tag] = tuple()
    deprecated: bool = False

    def __post_init__(self) -> None:
        if self.method.lower() not in self.supported_methods:
            raise ValueError(f'Unsupported method {self.method.lower()}, supported methods: {self.supported_methods}')
        object.__setattr__(self, 'securities', tuple(self.securities))
        param_full_name_to_param = {}
        for p in self.parameters:
            key = f'{p.location}.{p.name}'
            if present_p := param_full_name_to_param.get(key):
                raise KeyError(f'Parameter "{p}" has same {p.location.value=} and {p.name=} as "{present_p}')
            param_full_name_to_param[key] = p

        object.__setattr__(self, 'parameters', tuple(self.parameters))
        object.__setattr__(self, 'code_to_response', frozendict(self.code_to_response))
        object.__setattr__(self, 'tags', tuple(self.tags))

    def get_key(self) -> str:
        return self.method.lower()

    @property
    def spec_path(self) -> tuple[str, ...]:
        return 'paths', self.path

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        result: set[SpecResource] = set(self.securities) | set(self.parameters)
        result |= {t for t in self.tags if isinstance(t, Tag)}
        result.update(self.code_to_response.values())
        if self.request_body:
            result.add(self.request_body)
        return frozenset(result)

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result: dict[str, Jsonable] = {'operationId': self.operation_id}
        result |= {
            'tags': tuple(dependency_to_ref[t] if isinstance(t, Tag) else t for t in self.tags)} if self.tags else {}
        result |= {'summary': self.summary} if self.summary else {}
        result |= {'description': self.description} if self.description else {}
        result |= {'deprecated': self.deprecated} if self.deprecated else {}
        # result |= {'externalDocs': ...}  # TODO: implement external docs
        result |= {'security': tuple(dependency_to_ref[s] for s in self.securities)} if self.securities else {}
        result |= {'parameters': tuple(dependency_to_ref[p] for p in self.parameters)} if self.parameters else {}
        result |= {'requestBody': dependency_to_ref[self.request_body]} if self.request_body else {}
        if code_to_response := self.code_to_response:
            result['responses'] = {code: dependency_to_ref[response] for code, response in code_to_response.items()}
        # result |= {'callbacks': ...}  # TODO: implement callbacks
        # result |= {'servers': ...}  # TODO: implement servers
        return frozendict(result)

    def __repr__(self):
        parts = [f'path={self.path!r}', f'method={self.method!r}', f'operation_id={self.operation_id!r}']
        parts += [f'securities={self.securities!r}'] if self.securities else []
        parts += [f'parameters={list(self.parameters)!r}'] if self.parameters else []
        parts += [f'request_body={self.request_body!r}'] if self.request_body is not None else []
        if self.code_to_response:
            parts += [f'code_to_response={ {str(code): resp for code, resp in self.code_to_response.items() } }']
        parts += [f'summary={self.summary!r}'] if self.summary else []
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'tags={self.tags!r}'] if self.tags else []
        parts += [f'deprecated={self.deprecated!r}'] if self.deprecated else []
        return f'{self.__class__.__name__}({", ".join(parts)})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import json
from typing import Any

from frozendict import deepfreeze
from init_helpers import custom_dumps, try_extract_type_notes, AnyType


def freeze(val: Any):
    # Spike, TODO: think over it
    return deepfreeze(json.loads(custom_dumps(val)))


def repr_type(schema: AnyType) -> str:
    schema, notes = try_extract_type_notes(schema)
    if notes:
        return f'Annotated[{schema.__name__}, {", ".join(map(repr, notes))}]'
    return schema.__name__ if isinstance(schema, type) else str(schema)


#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from enum import StrEnum


class ParameterStyle(StrEnum):
    """
    Example with values
    string = "blue"
    array  = ["blue","black","brown"]
    object = {"R": 100, "G": 200, "B": 150}
    """

    matrix = 'matrix'
    # explode=False     ;color      ;color=blue     ;color=blue,black,brown                 ;color=R,100,G,200,B,150
    # explode=True      ;color      ;color=blue     ;color=blue;color=black;color=brown     ;R=100;G=200;B=150
    label = 'label'
    # explode=False     .           .blue           .blue.black.brown                       .R.100.G.200.B.150
    # explode=True      .           .blue           .blue.black.brown                       .R=100.G=200.B=150
    form = 'form'
    # explode=False     color=      color=blue      color=blue,black,brown                  color=R,100,G,200,B,150
    # explode=True      color=      color=blue      color=blue&color=black&color=brown      R=100&G=200&B=150
    simple = 'simple'
    # explode=False     n/a         blue            blue,black,brown                        R,100,G,200,B,150
    # explode=True      n/a         blue            blue,black,brown                        R=100,G=200,B=150
    spaceDelimited = 'spaceDelimited'
    # explode=False     n/a         n/a   	        blue%20black%20brown                    R%20100%20G%20200%20B%20150
    pipeDelimited = 'pipeDelimited'
    # explode=False     n/a         n/a             blue|black|brown                        R|100|G|200|B|150
    deepObject = 'deepObject'
    # explode=True      n/a         n/a             n/a                         color[R]=100&color[G]=200&color[B]=150



# #  Copyright (C) 2025
# #  ABM, Moscow
# #
# #  UNPUBLISHED PROPRIETARY MATERIAL.
# #  ALL RIGHTS RESERVED.
# #
# #  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
# from dataclasses import dataclass, field
#
# from init_helpers import Jsonable
#
# from .body_part_description import BodyPartDescription
# from .spec_resource import SpecResource
#
#
# @dataclass(frozen=True, slots=True)
# class Content(SpecResource):
#     schema: type
#     encoding: dict[str, BodyPartDescription] = field(default_factory=dict)
#
#     def get_spec_dependencies(self) -> set['SpecResource']:
#         raise NotImplementedError
#
#     def get_spec_dict(self) -> dict[str, Jsonable]:
#         raise NotImplementedError



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field
from typing import Any, ClassVar

from http_tools import IncomingRequest

from .parameter_location import ParameterLocation
from .spec_parameter import SpecParameter


@dataclass(frozen=True, slots=True, kw_only=True)
class PathParameter(SpecParameter):
    is_optional: bool = field(init=False, default=False)
    location: ClassVar[ParameterLocation] = ParameterLocation.path

    def _get(self, incoming_request: IncomingRequest) -> Any:
        return incoming_request.path_match_key_to_value[self.name]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors:
#  Mike Orlov <m.orlov@techokert.ru>
import json
import logging
import pathlib
from dataclasses import dataclass
from typing import Callable, Any

import yarl
from http_tools import HttpServer
from http_tools.answer import FileAnswer, RedirectAnswer
from http_tools.mime_types import ContentType
from http_tools.request import IncomingRequest
from init_helpers import custom_dumps, raise_if

from .endpoint import Endpoint
from openapi_tools.spec.spec_generator import SpecGenerator


logger = logging.getLogger(__name__)


def _return_empty_dict(*_: Any, **__: Any):
    return {}


class OpenApiServer:
    @dataclass
    class Context(SpecGenerator.Context):
        http_server: HttpServer

    @dataclass
    class Config(SpecGenerator.Config):
        write_spec_to_file: str = ""  # empty means skip
        http_prefix: str = '/.docs/'
        spec_http_path: str = 'openapi'  # empty means disable
        ui_http_path: str = 'ui'  # empty means disable

        def __post_init__(self):
            if prefix := self.http_prefix:
                raise_if(not prefix[0] == '/', ValueError(f"Http prefix MUST start with '/', got: {prefix}"))
                raise_if(not prefix[-1] == '/', ValueError(f"Http prefix MUST end with '/', got: {prefix}"))
            if spec_path := self.spec_http_path:
                raise_if('/' in spec_path, NotImplementedError(f"spec_http_path with '/' unsupported: {spec_path}"))
            if ui_path := self.ui_http_path:
                raise_if('/' in ui_path, NotImplementedError(f"ui_http_path with '/' unsupported: {ui_path}"))

    @property
    def _spec_path(self) -> str | None:
        if self.config.spec_http_path:
            return yarl.URL(self.config.http_prefix).join(yarl.URL(self.config.spec_http_path)).path

    @property
    def json_spec_path(self) -> str | None:
        if self._spec_path:
            return f'{self._spec_path}.json'

    @property
    def yaml_spec_paths(self) -> tuple[str, ...] | None:
        if self._spec_path:
            return f'{self._spec_path}.yml', f'{self._spec_path}.yaml'

    @property
    def ui_path(self) -> str | None:
        if self.config.ui_http_path:
            return yarl.URL(self.config.http_prefix).join(yarl.URL(self.config.ui_http_path)).path

    def __init__(
            self, config: Config, context: Context,
            spec_generator_factory: Callable[[Config, Context], SpecGenerator] = SpecGenerator
    ):
        self.context = context
        self.config = config
        self.spec_generator = spec_generator_factory(config, context)
        if self._spec_path:
            http_server = self.context.http_server
            http_server.register_handler(self._spec_path, self.get_spec_json)  # compat
            http_server.register_handler(self.json_spec_path, self.get_spec_json)
            for path in self.yaml_spec_paths:
                http_server.register_handler(path, self.get_spec_yaml)
            if self.ui_path:
                full_ui_path = self.ui_path + '/'
                http_server.register_handler(full_ui_path, self.get_ui)
                http_server.register_handler(self.ui_path, lambda _: RedirectAnswer(full_ui_path))
                ui_sub_path = yarl.URL(full_ui_path).join(yarl.URL('{file_name}')).path
                http_server.register_handler(ui_sub_path, self.get_ui)
                swagger_url = f'http://{http_server.config.interface}:{http_server.config.port}{full_ui_path}'
                logger.info(f'Swagger UI available at {swagger_url}')
        elif self.ui_path:
            logger.warning('Swagger UI disabled due to disabled spec endpoint')

    async def get_ui(self, request: IncomingRequest) -> FileAnswer:
        file_name = request.key_value_params.get('file_name', 'index.html')
        target = pathlib.Path(__file__).parent.parent / 'swagger-ui' / file_name
        if file_name.endswith('.html'):
            content = target.read_text()
            content_type = ContentType.HTML
        elif file_name.endswith('.css'):
            content = target.read_text()
            content_type = ContentType.Css
        elif file_name.endswith('.js'):
            content = target.read_text()
            content_type = ContentType.JavaScript
        elif file_name.endswith('.png'):
            content = target.read_bytes()
            content_type = ContentType.PNG
        else:
            raise TypeError(f"Unsupported file type: {file_name}")
        if file_name == 'swagger-initializer.js':
            content = content.replace(
                'https://petstore.swagger.io/v2/swagger.json', f'../{self.config.spec_http_path}.yaml')
        return FileAnswer(content, content_type=content_type)

    async def get_spec_json(self, _: IncomingRequest) -> FileAnswer:
        return FileAnswer(self.spec_generator.get_spec_json(), file_name='openapi.json', content_type=ContentType.Json)

    async def get_spec_yaml(self, _: IncomingRequest) -> FileAnswer:
        return FileAnswer(self.spec_generator.get_spec_yaml(), file_name='openapi.yaml', content_type=ContentType.Yaml)

    def _write_spec_to_file(self):
        spec = self.spec_generator.get_spec_dict()
        if self.config.write_spec_to_file:
            json_res = custom_dumps(spec, indent=2)
            with open(self.config.write_spec_to_file, "w") as f:
                f.write(json_res)

    def register_endpoint(self, rpc_endpoint: Endpoint) -> None:
        self.register(rpc_endpoint)

    def register(self, *rpc_endpoints: Endpoint) -> None:
        for endpoint in rpc_endpoints:
            self.context.http_server.register_handler(endpoint.path, endpoint.gen_handler(), [endpoint.method])
            self.spec_generator.register(endpoint.as_operation())
            self._write_spec_to_file()



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import http_tools
from init_helpers import try_extract_type_notes
from more_itertools import first

import openapi_tools.spec


def get_response_from_answer(answer: type[http_tools.Answer]) -> openapi_tools.spec.Response:
    answer, notes = try_extract_type_notes(answer)
    answer_examples = [note for note in notes if isinstance(note, openapi_tools.spec.AnswerExample)]
    content_type_to_media_type = {}
    if content_type := answer.get_class_content_type():
        content_type_to_media_type = {
            content_type: openapi_tools.spec.MediaType(answer.get_class_payload_type(), answer_examples)}
    return openapi_tools.spec.Response(
        description=first((note for note in notes if isinstance(note, str)), default=''),
        content_type_to_media_type=content_type_to_media_type, key=answer.__name__
    )
